library(tidyverse)
library(tidymodels)
library(vip)
library(pROC)
library(DALEX)
library(patchwork)

tidymodels_prefer(quiet = TRUE)
theme_set(theme_minimal())

# Input data
df <- read_rds("df_biology_simplified_grades.Rds")

# Create training/test data set
set.seed(123)

data_split <- initial_split(df, prop = 3/4, strata = graduation)
train_data <- training(data_split)
test_data <- testing(data_split)

# Logistic regression
# Remove the `id` column because glm cannot handle it
train_data_no_id <- train_data %>% select(-id)
test_data_no_id <- test_data %>% select(-id)

lm_fit <- glm(graduation ~ ., data = train_data_no_id, family = "binomial")
lm_pred_train <- predict(lm_fit, newdata = train_data_no_id, type = "response")

df_lm_pred <- tibble(Yes = 1 - lm_pred_train, 
                     No = lm_pred_train, 
                     graduation = train_data_no_id$graduation)

lm_roc_train <- df_lm_pred %>% 
  roc(response = graduation, predictor = Yes)

get_optimal_threshold <- function(roc) {
  tibble(TPR = roc$sensitivities, TNR = roc$specificities, threshold = roc$thresholds) %>% 
    mutate(diff = abs(TPR - TNR)) %>% 
    arrange(diff) %>% 
    slice(1) %>% 
    pull(threshold)
} 

lm_threshold <- get_optimal_threshold(lm_roc_train)

lm_pred_test <- predict(lm_fit, newdata = test_data_no_id, type = "response")

compute_metrics <- function(df, threshold) {
  df <- df %>% 
    mutate(pred_class = factor(ifelse(Yes > threshold, "Yes", "No"), levels = c("Yes", "No")))
  
  df %>% 
    roc_auc(truth = graduation, Yes) %>% 
    bind_rows(df %>% accuracy(truth = graduation, estimate = pred_class)) %>% 
    bind_rows(df %>% sensitivity(truth = graduation, estimate = pred_class)) %>% 
    bind_rows(df %>% specificity(truth = graduation, estimate = pred_class)) %>% 
    bind_rows(df %>% f_meas(truth = graduation, estimate = pred_class)) %>% 
    select(-.estimator)
}

df_lm_pred <- tibble(Yes = 1 - lm_pred_test,
                     No = lm_pred_test,
                     graduation = test_data_no_id$graduation)

compute_metrics(df_lm_pred, lm_threshold)

# Random forest
set.seed(123)
cv_folds <- vfold_cv(train_data, strata = graduation)

# Recipe for modeling
rf_rec <- recipe(graduation ~ ., data = train_data) %>% 
  update_role(id, new_role = "id")

# Model specification
rf_spec <- rand_forest(
  mtry = tune(),
  trees = 1000,
  min_n = tune()
) %>% 
  set_mode("classification") %>% 
  set_engine("ranger", importance = "permutation")

# Random forest workflow
rf_wf <- workflow() %>% 
  add_recipe(rf_rec) %>% 
  add_model(rf_spec)

# Tune grid for random forest parameters
set.seed(1234)
rf_grid <- grid_max_entropy(
  min_n(),
  finalize(mtry(), train_data),
  size = 30
)

doParallel::registerDoParallel()

set.seed(345)
rf_res <- tune_grid(
  rf_wf,
  resamples = cv_folds,
  grid = rf_grid
)

rf_best_auc <- select_best(rf_res, "roc_auc")

final_rf_wf <- rf_wf %>% 
  finalize_workflow(rf_best_auc)

final_rf_model <- final_rf_wf %>% 
  fit(data = train_data)

rf_pred_train <- predict(final_rf_model, train_data, type = "prob") %>% 
  bind_cols(train_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

rf_roc_train <- rf_pred_train %>% 
  roc(response = graduation, predictor = Yes)

rf_threshold <- get_optimal_threshold(rf_roc_train)

rf_pred_test <- predict(final_rf_model, test_data, type = "prob") %>% 
  bind_cols(test_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

compute_metrics(rf_pred_test, rf_threshold)

# XGBoost

xgb_rec <- recipe(graduation ~ ., data = train_data) %>% 
  update_role(id, new_role = "id") %>% 
  step_dummy(all_nominal_predictors(), one_hot = TRUE) %>% 
  step_novel(all_nominal_predictors())

xgb_spec <- boost_tree(
  trees = 1000,
  tree_depth = tune(),
  min_n = tune(),
  loss_reduction = tune(),
  sample_size = tune(),
  mtry = tune(),
  learn_rate = tune()
) %>% 
  set_engine("xgboost") %>% 
  set_mode("classification")

xgb_wf <- workflow() %>%
  add_recipe(xgb_rec) %>%
  add_model(xgb_spec)

xgb_grid <- grid_max_entropy(
  tree_depth(),
  min_n(),
  loss_reduction(),
  sample_size = sample_prop(),
  finalize(mtry(), train_data),
  learn_rate(),
  size = 30
)

doParallel::registerDoParallel()

set.seed(345)
xgb_res <- tune_grid(
  xgb_wf,
  resamples = cv_folds,
  grid = xgb_grid
)

xgb_best_auc <- select_best(xgb_res, "roc_auc")

final_xgb_wf <- xgb_wf %>% 
  finalize_workflow(xgb_best_auc)

tidymodels_prefer(quiet = TRUE)

final_xgb_model <- final_xgb_wf %>% 
  fit(data = train_data)

xgb_pred_train <- predict(final_xgb_model, train_data, type = "prob") %>% 
  bind_cols(train_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

xgb_roc_train <- xgb_pred_train %>% 
  roc(response = graduation, predictor = Yes)

xgb_threshold <- get_optimal_threshold(xgb_roc_train)

xgb_pred_test <- predict(final_xgb_model, test_data, type = "prob") %>% 
  bind_cols(test_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

compute_metrics(xgb_pred_test, xgb_threshold)

# KNN

knn_rec <- recipe(graduation ~ ., data = train_data) %>% 
  update_role(id, new_role = "id") %>% 
  step_dummy(all_nominal_predictors(), one_hot = TRUE) %>% 
  step_novel(all_nominal_predictors())

knn_spec <- nearest_neighbor(
  neighbors = tune()
) %>% 
  set_engine("kknn") %>% 
  set_mode("classification")

knn_wf <- workflow() %>%
  add_recipe(knn_rec) %>%
  add_model(knn_spec)

knn_grid <- tibble(neighbors = seq(from = 1, to = 30, by = 1))

doParallel::registerDoParallel()

set.seed(345)
knn_res <- tune_grid(
  knn_wf,
  resamples = cv_folds,
  grid = knn_grid
)

knn_best_auc <- select_best(knn_res, "roc_auc")

final_knn_wf <- knn_wf %>% 
  finalize_workflow(knn_best_auc)

final_knn_model <- final_knn_wf %>% 
  fit(data = train_data)

knn_pred_train <- predict(final_knn_model, train_data, type = "prob") %>% 
  bind_cols(train_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

knn_roc_train <- knn_pred_train %>% 
  roc(response = graduation, predictor = Yes)

knn_threshold <- get_optimal_threshold(knn_roc_train)

knn_pred_test <- predict(final_knn_model, test_data, type = "prob") %>% 
  bind_cols(test_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

compute_metrics(knn_pred_test, knn_threshold)

# Neural network

nn_spec <-
  mlp(hidden_units = tune(), 
      penalty = tune(), 
      epochs = tune()) %>%
  set_mode("classification") %>%
  set_engine("nnet")

nn_rec <- recipe(graduation ~ ., data = train_data) %>% 
  update_role(id, new_role = "id") %>% 
  step_dummy(all_nominal_predictors(), one_hot = TRUE) %>% 
  step_zv() %>% 
  step_normalize(all_predictors())

nn_wf <- workflow() %>%
  add_recipe(nn_rec) %>%
  add_model(nn_spec)

nn_grid <- grid_max_entropy(
  hidden_units(),
  penalty(),
  epochs(),
  size = 30
)

doParallel::registerDoParallel()

set.seed(345)
nn_res <- tune_grid(
  nn_wf,
  resamples = cv_folds,
  grid = nn_grid
)

nn_best_auc <- select_best(nn_res, "roc_auc")

final_nn_wf <- nn_wf %>% 
  finalize_workflow(nn_best_auc)

final_nn_model <- final_nn_wf %>% 
  fit(data = train_data)

nn_pred_train <- predict(final_nn_model, train_data, type = "prob") %>% 
  bind_cols(train_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

nn_roc_train <- nn_pred_train %>% 
  roc(response = graduation, predictor = Yes)

nn_threshold <- get_optimal_threshold(nn_roc_train)

nn_pred_test <- predict(final_nn_model, test_data, type = "prob") %>% 
  bind_cols(test_data %>% dplyr::select(graduation)) %>% 
  rename(Yes = .pred_Yes, No = .pred_No)

compute_metrics(nn_pred_test, nn_threshold)

# Explanable ML using DALEX
explanier_rf <- DALEX::explain(
  model = final_rf_model %>% extract_fit_parsnip(),
  data = test_data %>% select(-graduation),
  y = ifelse(test_data %>% pull(graduation) == "Yes", 1, 0),
  predict_function = function(model, newdata)  {
    result <- predict(model, newdata, type = "prob")
    return(result$.pred_Yes)
  },
  label = "Random Forest"
)

# Negative instance
df_prediction_error <- rf_pred_test %>% 
  mutate(id = test_data %>% pull(id)) %>% 
  select(id, everything()) %>% 
  mutate(graduation = as.character(graduation)) %>% 
  mutate(graduation_numeric = ifelse(graduation == "Yes", 1, 0)) %>% 
  mutate(prediction_error = graduation_numeric - Yes)

df_prediction_error <- left_join(test_data, df_prediction_error) 

df_prediction_error <- df_prediction_error %>% 
  mutate(pred = ifelse(Yes > rf_threshold, "Yes", "No")) %>% 
  mutate(pred_correct = ifelse(pred == graduation, "Yes", "No"))

negative_data <- df_prediction_error %>% 
  filter(id == 606)

breakdown_negative <- DALEX::predict_parts(explainer = explanier_rf,
                                           new_observation = negative_data,
                                           type = "break_down") %>% 
  plot(title = NULL) + 
  labs(subtitle = NULL) # Disabling the subtitle does not work

breakdown_negative <- breakdown_negative +
  geom_hline(yintercept = rf_threshold, linetype = "dashed", color = "red", alpha = 0.5)

# Positive instance
positive_data <- df_prediction_error %>% 
  filter(id == 379)

breakdown_positive <- DALEX::predict_parts(explainer = explanier_rf,
                                           new_observation = positive_data,
                                           type = "break_down") %>% 
  plot(title = NULL) + 
  labs(subtitle = NULL) # Disabling the subtitle does not work

breakdown_positive <- breakdown_positive +
  geom_hline(yintercept = rf_threshold, linetype = "dashed", color = "red", alpha = 0.5)

breakdown_negative + breakdown_positive +
  plot_annotation(tag_levels = 'A')

sessionInfo()
